@@ -52,7 +52,7 @@ enum
 #define BOM_UTF32_LE "\xff\xfe\x00\x00"
 #define BOM_UTF32_BE "\x00\x00\xfe\xff"
 
-typedef struct _JSONDecoder {
+typedef struct _Decoder {
 	Py_UNICODE *start;
 	Py_UNICODE *end;
 	Py_UNICODE *index;
@@ -62,7 +62,7 @@ typedef struct _JSONDecoder {
 	size_t stringparse_buffer_size;
 	
 	unsigned int got_root: 1;
-} JSONDecoder;
+} Decoder;
 
 typedef enum
 {
@@ -82,16 +82,16 @@ typedef enum
 
 static PyObject *ReadError;
 
-static PyObject *read_string (JSONDecoder *decoder);
-static PyObject *read_number (JSONDecoder *decoder);
-static PyObject *read_array (JSONDecoder *decoder);
-static PyObject *read_object (JSONDecoder *decoder);
-static PyObject *json_read (JSONDecoder *decoder);
+static PyObject *read_string (Decoder *decoder);
+static PyObject *read_number (Decoder *decoder);
+static PyObject *read_array (Decoder *decoder);
+static PyObject *read_object (Decoder *decoder);
+static PyObject *json_read (Decoder *decoder);
 /* }}} */
 
 /* serializer declarations {{{ */
-typedef struct _JSONEncoder JSONEncoder;
-struct _JSONEncoder
+typedef struct _Encoder Encoder;
+struct _Encoder
 {
 	/* Pulled from the current interpreter to avoid errors when used
 	 * with sub-interpreters.
@@ -106,8 +106,8 @@ struct _JSONEncoder
 	int coerce_keys;
 	PyObject *on_unknown;
 	
-	int (*append_ascii) (JSONEncoder *, const char *, const size_t);
-	int (*append_unicode) (JSONEncoder *, const Py_UNICODE *, const size_t);
+	int (*append_ascii) (Encoder *, const char *, const size_t);
+	int (*append_unicode) (Encoder *, PyObject *);
 	
 	/* Constants, saved to avoid lookup later */
 	PyObject *true_str;
@@ -120,20 +120,20 @@ struct _JSONEncoder
 	PyObject *colon;
 };
 
-typedef struct _JSONBufferEncoder
+typedef struct _BufferEncoder
 {
-	JSONEncoder encoder;
+	Encoder encoder;
 	Py_UNICODE *buffer;
 	size_t buffer_size;
 	size_t buffer_max_size;
-} JSONBufferEncoder;
+} BufferEncoder;
 
-typedef struct _JSONStreamEncoder
+typedef struct _StreamEncoder
 {
-	JSONEncoder encoder;
+	Encoder encoder;
 	PyObject *stream;
 	char *encoding;
-} JSONStreamEncoder;
+} StreamEncoder;
 
 static const char *hexdigit = "0123456789abcdef";
 #define INITIAL_BUFFER_SIZE 32
@@ -143,57 +143,52 @@ static PyObject *UnknownSerializerError;
 
 /* Functions for writing actual bytes to a buffer or stream {{{ */
 static int
-encoder_buffer_append_ascii (JSONEncoder *encoder,
+buffer_encoder_append_ascii (Encoder *encoder,
                              const char *text,
                              const size_t len);
 
 static int
-encoder_buffer_append_unicode (JSONEncoder *encoder,
-                               const Py_UNICODE *text,
-                               const size_t len);
+buffer_encoder_append_unicode (Encoder *encoder,
+                               PyObject *text);
 
 static int
-encoder_stream_append_ascii (JSONEncoder *encoder,
+stream_encoder_append_ascii (Encoder *encoder,
                              const char *text,
                              const size_t len);
 
 static int
-encoder_stream_append_unicode (JSONEncoder *encoder,
-                               const Py_UNICODE *text,
-                               const size_t len);
+stream_encoder_append_unicode (Encoder *encoder,
+                               PyObject *text);
 
 static int
-encoder_append_const (JSONEncoder *encoder, const char text[]);
+encoder_append_string (Encoder *encoder, PyObject *text);
 
 static int
-encoder_append_string (JSONEncoder *encoder, PyObject *text);
-
-static int
-encoder_buffer_resize (JSONBufferEncoder *encoder, size_t delta);
+buffer_encoder_resize (BufferEncoder *encoder, size_t delta);
 /* }}} */
 
 
 static PyObject *
-unicode_from_ascii (const char *value);
+ascii_constant (const char *value, int len);
 
 static int
-write_object (JSONEncoder *encoder, PyObject *object, int indent_level,
+write_object (Encoder *encoder, PyObject *object, int indent_level,
               int in_unknown_hook);
 
 static int
-write_iterable (JSONEncoder *encoder, PyObject *iterable, int indent_level);
+write_iterable (Encoder *encoder, PyObject *iterable, int indent_level);
 
 static int
-write_mapping (JSONEncoder *encoder, PyObject *mapping, int indent_level);
+write_mapping (Encoder *encoder, PyObject *mapping, int indent_level);
 
 static PyObject *
-write_basic (JSONEncoder *encoder, PyObject *value);
+write_basic (Encoder *encoder, PyObject *value);
 
 static PyObject *
-write_string (JSONEncoder *encoder, PyObject *string);
+write_string (Encoder *encoder, PyObject *string);
 
 static PyObject *
-write_unicode (JSONEncoder *encoder, PyObject *unicode);
+write_unicode (Encoder *encoder, PyObject *unicode);
 
 static PyObject *
 unicode_to_unicode (PyObject *unicode);
@@ -238,7 +233,7 @@ next_power_2 (size_t start, size_t min)
 
 /* parser {{{ */
 static void
-skip_spaces (JSONDecoder *decoder)
+skip_spaces (Decoder *decoder)
 {
 	/* Don't use Py_UNICODE_ISSPACE, because it returns TRUE for
 	 * codepoints that are not valid JSON whitespace.
@@ -254,7 +249,7 @@ skip_spaces (JSONDecoder *decoder)
 }
 
 static int
-parser_find_next_value (JSONDecoder *decoder)
+parser_find_next_value (Decoder *decoder)
 {
 	/* Return codes:
 	 * 
@@ -339,7 +334,7 @@ count_row_column (Py_UNICODE *start, Py_UNICODE *pos, unsigned long *offset,
 }
 
 static void
-set_error (JSONDecoder *decoder, Py_UNICODE *position, PyObject *description,
+set_error (Decoder *decoder, Py_UNICODE *position, PyObject *description,
            PyObject *description_args)
 {
 	const char *tmpl = "JSON parsing error at line %d, column %d"
@@ -382,7 +377,7 @@ set_error (JSONDecoder *decoder, Py_UNICODE *position, PyObject *description,
 }
 
 static void
-set_error_simple (JSONDecoder *decoder, Py_UNICODE *position,
+set_error_simple (Decoder *decoder, Py_UNICODE *position,
                   const char *description)
 {
 	PyObject *desc_obj;
@@ -395,7 +390,7 @@ set_error_simple (JSONDecoder *decoder, Py_UNICODE *position,
 }
 
 static void
-set_error_unexpected (JSONDecoder *decoder, Py_UNICODE *position,
+set_error_unexpected (Decoder *decoder, Py_UNICODE *position,
                       const char *wanted)
 {
 	PyObject *err_str, *err_format_args;
@@ -433,7 +428,7 @@ set_error_unexpected (JSONDecoder *decoder, Py_UNICODE *position,
 
 /* Helper function to create a new decimal.Decimal object */
 static PyObject *
-make_Decimal (JSONDecoder *decoder, PyObject *string)
+make_Decimal (Decoder *decoder, PyObject *string)
 {
 	PyObject *args, *retval = NULL;
 	
@@ -446,7 +441,7 @@ make_Decimal (JSONDecoder *decoder, PyObject *string)
 }
 
 static PyObject *
-keyword_compare (JSONDecoder *decoder, const char *expected, size_t len,
+keyword_compare (Decoder *decoder, const char *expected, size_t len,
                  PyObject *retval)
 {
 	size_t ii, left;
@@ -480,7 +475,7 @@ read_4hex (Py_UNICODE *start, Py_UNICODE *retval_ptr)
 }
 
 static int
-read_unicode_escape (JSONDecoder *decoder, Py_UNICODE *string_start,
+read_unicode_escape (Decoder *decoder, Py_UNICODE *string_start,
                      Py_UNICODE *buffer, size_t *buffer_idx,
                      size_t *index_ptr, size_t max_char_count)
 {
@@ -569,52 +564,14 @@ read_unicode_escape (JSONDecoder *decoder, Py_UNICODE *string_start,
 }
 
 static PyObject *
-read_string (JSONDecoder *decoder)
+read_string_full (Decoder *decoder, Py_UNICODE *start, size_t max_char_count)
 {
 	PyObject *unicode;
 	int escaped = FALSE;
-	Py_UNICODE c, *buffer, *start;
-	size_t ii, max_char_count, buffer_idx;
-	
-	/* Start at 1 to skip first double quote. */
-	start = decoder->index + 1;
-	
-	/* Scan through for maximum character count, and to ensure the string
-	 * is terminated.
-	**/
-	for (ii = 0;; ii++)
-	{
-		c = start[ii];
-		if (c == 0)
-		{
-			set_error_simple (decoder, decoder->index,
-			                  "Unterminated string.");
-			return NULL;
-		}
-		
-		/* Check for illegal characters */
-		if (c < 0x20)
-		{
-			set_error_unexpected (decoder, start + ii, NULL);
-			return NULL;
-		}
-		
-		if (escaped)
-		{
-			/* Invalid escape codes will be caught
-			 * later.
-			**/
-			escaped = FALSE;
-		}
-		
-		else
-		{	if (c == '\\') escaped = TRUE;
-			else if (c == '"') break;
-		}
-	}
+	Py_UNICODE c, *buffer;
+	size_t ii, buffer_idx;
 	
 	/* Allocate enough to hold the worst case */
-	max_char_count = ii;
 	buffer = decoder->stringparse_buffer;
 	if (max_char_count > decoder->stringparse_buffer_size)
 	{
@@ -711,7 +668,77 @@ read_string (JSONDecoder *decoder)
 }
 
 static PyObject *
-read_number (JSONDecoder *decoder)
+read_string (Decoder *decoder)
+{
+	PyObject *unicode;
+	int escaped = FALSE, fancy = FALSE;
+	Py_UNICODE c, *start;
+	size_t ii;
+	
+	/* Start at 1 to skip first double quote. */
+	start = decoder->index + 1;
+	
+	/* Fast case for empty string */
+	if (start[0] == '"')
+	{
+		decoder->index = start + 1;
+		return PyUnicode_FromUnicode (NULL, 0);
+	}
+	
+	/* Scan through for maximum character count, and to ensure the string
+	 * is terminated.
+	**/
+	for (ii = 0;; ii++)
+	{
+		c = start[ii];
+		if (c == 0)
+		{
+			set_error_simple (decoder, decoder->index,
+			                  "Unterminated string.");
+			return NULL;
+		}
+		
+		/* Check for illegal characters */
+		if (c < 0x20)
+		{
+			set_error_unexpected (decoder, start + ii, NULL);
+			return NULL;
+		}
+		
+		if (escaped)
+		{
+			/* Invalid escape codes will be caught
+			 * later.
+			**/
+			escaped = FALSE;
+		}
+		
+		else
+		{	if (c == '\\')
+			{
+				fancy = TRUE;
+				escaped = TRUE;
+			}
+			else if (c == '"') break;
+		}
+	}
+	
+	if (fancy)
+	{
+		return read_string_full (decoder, start, ii);
+	}
+	
+	/* No fancy features, return the string directly */
+	unicode = PyUnicode_FromUnicode (start, ii);
+	if (unicode)
+	{
+		decoder->index = start + ii + 1;
+	}
+	return unicode;
+}
+
+static PyObject *
+read_number (Decoder *decoder)
 {
 	PyObject *object = NULL;
 	int is_float = FALSE, should_stop = FALSE, got_digit = FALSE,
@@ -806,7 +833,7 @@ read_number (JSONDecoder *decoder)
 }
 
 static int
-read_array_impl (PyObject *list, JSONDecoder *decoder)
+read_array_impl (PyObject *list, Decoder *decoder)
 {
 	Py_UNICODE *start, c;
 	ParseArrayState array_state = ARRAY_EMPTY;
@@ -873,7 +900,7 @@ read_array_impl (PyObject *list, JSONDecoder *decoder)
 }
 
 static PyObject *
-read_array (JSONDecoder *decoder)
+read_array (Decoder *decoder)
 {
 	PyObject *object = PyList_New (0);
 	
@@ -887,7 +914,7 @@ read_array (JSONDecoder *decoder)
 }
 
 static int
-read_object_impl (PyObject *object, JSONDecoder *decoder)
+read_object_impl (PyObject *object, Decoder *decoder)
 {
 	Py_UNICODE *start, c;
 	ParseObjectState object_state = OBJECT_EMPTY;
@@ -986,7 +1013,7 @@ read_object_impl (PyObject *object, JSONDecoder *decoder)
 }
 
 static PyObject *
-read_object (JSONDecoder *decoder)
+read_object (Decoder *decoder)
 {
 	PyObject *object = PyDict_New ();
 	
@@ -1000,7 +1027,7 @@ read_object (JSONDecoder *decoder)
 }
 
 static PyObject *
-json_read (JSONDecoder *decoder)
+json_read (Decoder *decoder)
 {
 	skip_spaces (decoder);
 	switch (*decoder->index)
@@ -1175,7 +1202,7 @@ static PyObject*
 _read_entry (PyObject *self, PyObject *args, PyObject *kwargs)
 {
 	PyObject *result = NULL, *unicode;
-	JSONDecoder decoder = {NULL};
+	Decoder decoder = {NULL};
 	
 	if (!parse_unicode_arg (args, kwargs, &unicode))
 		return NULL;
@@ -1219,14 +1246,14 @@ _read_entry (PyObject *self, PyObject *args, PyObject *kwargs)
 
 /* serializer {{{ */
 static int
-encoder_buffer_append_ascii (JSONEncoder *base_encoder,
+buffer_encoder_append_ascii (Encoder *base_encoder,
                              const char *text,
                              const size_t len)
 {
-	JSONBufferEncoder *encoder = (JSONBufferEncoder *) (base_encoder);
+	BufferEncoder *encoder = (BufferEncoder *) (base_encoder);
 	size_t ii;
 	
-	if (!encoder_buffer_resize (encoder, len))
+	if (!buffer_encoder_resize (encoder, len))
 		return FALSE;
 	for (ii = 0; ii < len; ii++)
 	{
@@ -1236,24 +1263,26 @@ encoder_buffer_append_ascii (JSONEncoder *base_encoder,
 }
 
 static int
-encoder_buffer_append_unicode (JSONEncoder *base_encoder,
-                               const Py_UNICODE *text,
-                               const size_t len)
+buffer_encoder_append_unicode (Encoder *base_encoder,
+                               PyObject *text)
 {
-	JSONBufferEncoder *encoder = (JSONBufferEncoder *) (base_encoder);
-	size_t ii;
+	size_t len;
+	Py_UNICODE *raw;
+	BufferEncoder *encoder = (BufferEncoder *) (base_encoder);
 	
-	if (!encoder_buffer_resize (encoder, len))
+	raw = PyUnicode_AS_UNICODE (text);
+	len = PyUnicode_GET_SIZE (text);
+	
+	if (!buffer_encoder_resize (encoder, len))
 		return FALSE;
-	for (ii = 0; ii < len; ii++)
-	{
-		encoder->buffer[encoder->buffer_size++] = text[ii];
-	}
+	memcpy (encoder->buffer + encoder->buffer_size, raw,
+	        len * sizeof (Py_UNICODE));
+	encoder->buffer_size += len;
 	return TRUE;
 }
 
 static int
-encoder_stream_append_common (JSONStreamEncoder *encoder, PyObject *encoded)
+stream_encoder_append_common (StreamEncoder *encoder, PyObject *encoded)
 {
 	int result;
 	if (!encoded)
@@ -1264,48 +1293,43 @@ encoder_stream_append_common (JSONStreamEncoder *encoder, PyObject *encoded)
 }
 
 static int
-encoder_stream_append_ascii (JSONEncoder *base_encoder,
+stream_encoder_append_ascii (Encoder *base_encoder,
                              const char *text,
                              const size_t len)
 {
-	JSONStreamEncoder *encoder = (JSONStreamEncoder *) (base_encoder);
+	StreamEncoder *encoder = (StreamEncoder *) (base_encoder);
 	PyObject *encoded;
 	if (encoder->encoding)
 		encoded = PyString_Encode (text, len, encoder->encoding, "strict");
 	else
 		encoded = PyUnicode_Decode (text, len, "ascii", "strict");
-	return encoder_stream_append_common (encoder, encoded);
+	return stream_encoder_append_common (encoder, encoded);
 }
 
 static int
-encoder_stream_append_unicode (JSONEncoder *base_encoder,
-                               const Py_UNICODE *text,
-                               const size_t len)
+stream_encoder_append_unicode (Encoder *base_encoder,
+                               PyObject *text)
 {
-	JSONStreamEncoder *encoder = (JSONStreamEncoder *) (base_encoder);
+	StreamEncoder *encoder = (StreamEncoder *) (base_encoder);
 	PyObject *encoded;
 	if (encoder->encoding)
-		encoded = PyUnicode_Encode (text, len, encoder->encoding, "strict");
+		encoded = PyUnicode_AsEncodedString (text, encoder->encoding, "strict");
 	else
-		encoded = PyUnicode_FromUnicode (text, len);
-	return encoder_stream_append_common (encoder, encoded);
+	{
+		encoded = text;
+		Py_INCREF (encoded);
+	}
+	return stream_encoder_append_common (encoder, encoded);
 }
 
 static int
-encoder_append_const (JSONEncoder *encoder, const char text[])
-{
-	return encoder->append_ascii (encoder, text, strlen (text));
-}
-
-static int
-encoder_append_string (JSONEncoder *encoder, PyObject *text)
+encoder_append_string (Encoder *encoder, PyObject *text)
 {
 	size_t len;
+	
 	if (PyUnicode_CheckExact (text))
 	{
-		Py_UNICODE *raw = PyUnicode_AS_UNICODE (text);
-		len = PyUnicode_GET_SIZE (text);
-		return encoder->append_unicode (encoder, raw, len);
+		return encoder->append_unicode (encoder, text);
 	}
 	if (PyString_CheckExact (text))
 	{
@@ -1319,9 +1343,15 @@ encoder_append_string (JSONEncoder *encoder, PyObject *text)
 }
 
 static int
-encoder_buffer_resize (JSONBufferEncoder *encoder, size_t delta)
+buffer_encoder_resize (BufferEncoder *encoder, size_t delta)
 {
 	size_t new_size;
+	Py_UNICODE *new_buf;
+	
+	new_size = encoder->buffer_size + delta;
+	if (encoder->buffer_max_size >= new_size)
+		return TRUE;
+		
 	if (!encoder->buffer)
 	{
 		new_size = (delta > INITIAL_BUFFER_SIZE? delta : INITIAL_BUFFER_SIZE);
@@ -1331,22 +1361,16 @@ encoder_buffer_resize (JSONBufferEncoder *encoder, size_t delta)
 		return TRUE;
 	}
 	
-	new_size = encoder->buffer_size + delta;
-	if (encoder->buffer_max_size < new_size)
+	new_size = next_power_2 (encoder->buffer_max_size, new_size);
+	new_buf = PyMem_Realloc (encoder->buffer,
+	                         sizeof (Py_UNICODE) * new_size);
+	if (!new_buf)
 	{
-		Py_UNICODE *new_buf;
-		new_size = next_power_2 (encoder->buffer_max_size, new_size);
-		new_buf = PyMem_Realloc (encoder->buffer,
-		                         sizeof (Py_UNICODE) * new_size);
-		if (!new_buf)
-		{
-			PyMem_Free (encoder->buffer);
-			return FALSE;
-		}
-		encoder->buffer = new_buf;
-		encoder->buffer_max_size = new_size;
-		return TRUE;
+		PyMem_Free (encoder->buffer);
+		return FALSE;
 	}
+	encoder->buffer = new_buf;
+	encoder->buffer_max_size = new_size;
 	return TRUE;
 }
 
@@ -1365,13 +1389,11 @@ set_unknown_serializer (PyObject *value)
 }
 
 static PyObject *
-unicode_from_ascii (const char *value)
+ascii_constant (const char *value, int len)
 {
-	PyObject *str, *retval;
-	str = PyString_FromString (value);
-	retval = PyUnicode_FromEncodedObject (str, "ascii", "strict");
-	Py_DECREF (str);
-	return retval;
+	if (len < 0)
+		len = strlen (value);
+	return PyUnicode_DecodeASCII (value, len, "strict");
 }
 
 static PyObject *
@@ -1398,10 +1420,10 @@ get_separators (PyObject *indent_string, int indent_level,
 {
 	if (indent_string == Py_None)
 	{
-		(*start_ptr) = PyString_FromStringAndSize (&start, 1);
+		(*start_ptr) = ascii_constant (&start, 1);
 		(*pre_value_ptr) = NULL;
-		(*post_value_ptr) = PyString_FromStringAndSize (",", 1);
-		(*end_ptr) = PyString_FromStringAndSize (&end, 1);
+		(*post_value_ptr) = ascii_constant (",", 1);
+		(*end_ptr) = ascii_constant (&end, 1);
 	}
 	else
 	{
@@ -1409,8 +1431,8 @@ get_separators (PyObject *indent_string, int indent_level,
 		char start_str[] = {0, '\n'};
 		start_str[0] = start;
 		
-		(*start_ptr) = PyString_FromStringAndSize (start_str, 2);
-		(*post_value_ptr) = PyString_FromStringAndSize (",\n", 2);
+		(*start_ptr) = ascii_constant (start_str, 2);
+		(*post_value_ptr) = ascii_constant (",\n", 2);
 		
 		indent = PySequence_Repeat (indent_string, indent_level + 1);
 		(*pre_value_ptr) = indent;
@@ -1425,7 +1447,7 @@ get_separators (PyObject *indent_string, int indent_level,
 }
 
 static PyObject *
-write_string (JSONEncoder *encoder, PyObject *string)
+write_string (Encoder *encoder, PyObject *string)
 {
 	PyObject *unicode, *retval;
 	int safe = TRUE;
@@ -1560,6 +1582,58 @@ unicode_to_unicode (PyObject *unicode)
 	return retval;
 }
 
+static char *
+escape_unichar (Py_UNICODE c, char *p)
+{
+	*p++ = '\\';
+	switch (c)
+	{
+		case 0x08: *p++ = 'b'; return p;
+		case 0x09: *p++ = 't'; return p;
+		case 0x0A: *p++ = 'n'; return p;
+		case 0x0C: *p++ = 'f'; return p;
+		case 0x0D: *p++ = 'r'; return p;
+		case 0x22: *p++ = '"'; return p;
+		case 0x2F: *p++ = '/'; return p;
+		case 0x5C: *p++ = '\\'; return p;
+		default: break;
+	}
+#ifdef Py_UNICODE_WIDE
+	if (c > 0xFFFF)
+	{
+		/* Separate into upper and lower surrogate pair */
+		Py_UNICODE reduced, upper, lower;
+		
+		reduced = c - 0x10000;
+		lower = (reduced & 0x3FF);
+		upper = (reduced >> 10);
+		
+		upper += 0xD800;
+		lower += 0xDC00;
+		
+		*p++ = 'u';
+		*p++ = hexdigit[(upper >> 12) & 0x0000000F];
+		*p++ = hexdigit[(upper >> 8) & 0x0000000F];
+		*p++ = hexdigit[(upper >> 4) & 0x0000000F];
+		*p++ = hexdigit[upper & 0x0000000F];
+		
+		*p++ = '\\';
+		*p++ = 'u';
+		*p++ = hexdigit[(lower >> 12) & 0x0000000F];
+		*p++ = hexdigit[(lower >> 8) & 0x0000000F];
+		*p++ = hexdigit[(lower >> 4) & 0x0000000F];
+		*p++ = hexdigit[lower & 0x0000000F];
+		return p;
+	}
+#endif
+	*p++ = 'u';
+	*p++ = hexdigit[(c >> 12) & 0x000F];
+	*p++ = hexdigit[(c >> 8) & 0x000F];
+	*p++ = hexdigit[(c >> 4) & 0x000F];
+	*p++ = hexdigit[c & 0x000F];
+	return p;
+}
+
 static PyObject *
 unicode_to_ascii (PyObject *unicode)
 {
@@ -1626,77 +1700,17 @@ unicode_to_ascii (PyObject *unicode)
 	for (ii = 0; ii < old_buffer_size; ii++)
 	{
 		Py_UNICODE c = old_buffer[ii];
-		if (c == 0x08)
-			*p++ = '\\', *p++ = 'b';
-		else if (c == 0x09)
-			*p++ = '\\', *p++ = 't';
-		else if (c == 0x0A)
-			*p++ = '\\', *p++ = 'n';
-		else if (c == 0x0C)
-			*p++ = '\\', *p++ = 'f';
-		else if (c == 0x0D)
-			*p++ = '\\', *p++ = 'r';
-		else if (c == 0x22)
-			*p++ = '\\', *p++ = '"';
-		else if (c == 0x2F)
-			*p++ = '\\', *p++ = '/';
-		else if (c == 0x5C)
-			*p++ = '\\', *p++ = '\\';
-		else if (c <= 0x1F)
-		{
-			*p++ = '\\';
-			*p++ = 'u';
-			*p++ = '0';
-			*p++ = '0';
-			*p++ = hexdigit[(c >> 4) & 0x0000000F];
-			*p++ = hexdigit[c & 0x0000000F];
-		}
-#ifdef Py_UNICODE_WIDE
-		else if (c > 0xFFFF)
-		{
-			/* Separate into upper and lower surrogate pair */
-			Py_UNICODE reduced, upper, lower;
-			
-			reduced = c - 0x10000;
-			lower = (reduced & 0x3FF);
-			upper = (reduced >> 10);
-			
-			upper += 0xD800;
-			lower += 0xDC00;
-			
-			*p++ = '\\';
-			*p++ = 'u';
-			*p++ = hexdigit[(upper >> 12) & 0x0000000F];
-			*p++ = hexdigit[(upper >> 8) & 0x0000000F];
-			*p++ = hexdigit[(upper >> 4) & 0x0000000F];
-			*p++ = hexdigit[upper & 0x0000000F];
-			
-			*p++ = '\\';
-			*p++ = 'u';
-			*p++ = hexdigit[(lower >> 12) & 0x0000000F];
-			*p++ = hexdigit[(lower >> 8) & 0x0000000F];
-			*p++ = hexdigit[(lower >> 4) & 0x0000000F];
-			*p++ = hexdigit[lower & 0x0000000F];
-		}
-#endif
-		else if (c > 0x7E)
-		{
-			*p++ = '\\';
-			*p++ = 'u';
-			*p++ = hexdigit[(c >> 12) & 0x000F];
-			*p++ = hexdigit[(c >> 8) & 0x000F];
-			*p++ = hexdigit[(c >> 4) & 0x000F];
-			*p++ = hexdigit[c & 0x000F];
-		}
-		else
+		if (c > 0x1F && c <= 0x7E && c != '\\' && c != '"' && c != '/')
 			*p++ = (char) (c);
+		else
+			p = escape_unichar (c, p);
 	}
 	*p++ = '"';
 	return retval;
 }
 
 static PyObject *
-write_unicode (JSONEncoder *encoder, PyObject *unicode)
+write_unicode (Encoder *encoder, PyObject *unicode)
 {
 	PyObject *retval;
 	int safe = TRUE;
@@ -1722,15 +1736,16 @@ write_unicode (JSONEncoder *encoder, PyObject *unicode)
 	
 	if (safe)
 	{
-		PyObject *seq = NULL, *sep = NULL;
-		seq = Py_BuildValue ("(OOO)",
-		                     encoder->quote,
-		                     unicode,
-		                     encoder->quote);
-		sep = PyUnicode_FromUnicode (NULL, 0);
-		retval = PyUnicode_Join (sep, seq);
-		Py_XDECREF (seq);
-		Py_XDECREF (sep);
+		PyUnicodeObject *u_retval;
+		
+		if (!(retval = PyUnicode_FromUnicode (NULL, str_len + 2)))
+			return NULL;
+			
+		u_retval = (PyUnicodeObject*) retval;
+		Py_UNICODE_COPY (u_retval->str + 1, buffer, str_len);
+		u_retval->str[0] = '"';
+		u_retval->str[str_len + 1] = '"';
+		
 		return retval;
 	}
 	
@@ -1774,7 +1789,7 @@ write_unicode (JSONEncoder *encoder, PyObject *unicode)
 }
 
 static int
-write_sequence_impl (JSONEncoder *encoder, PyObject *seq,
+write_sequence_impl (Encoder *encoder, PyObject *seq,
                      PyObject *start, PyObject *end,
                      PyObject *pre_value, PyObject *post_value,
                      int indent_level)
@@ -1809,7 +1824,7 @@ write_sequence_impl (JSONEncoder *encoder, PyObject *seq,
 }
 
 static int
-write_iterable (JSONEncoder *encoder, PyObject *iter, int indent_level)
+write_iterable (Encoder *encoder, PyObject *iter, int indent_level)
 {
 	PyObject *sequence;
 	PyObject *start, *end, *pre, *post;
@@ -1832,7 +1847,7 @@ write_iterable (JSONEncoder *encoder, PyObject *iter, int indent_level)
 	{
 		Py_DECREF (sequence);
 		Py_ReprLeave (iter);
-		return encoder_append_const (encoder, "[]");
+		return encoder->append_ascii (encoder, "[]", 2);
 	}
 	
 	/* Build separator strings */
@@ -1854,7 +1869,7 @@ write_iterable (JSONEncoder *encoder, PyObject *iter, int indent_level)
 }
 
 static int
-mapping_process_key (JSONEncoder *encoder, PyObject *key, PyObject **key_ptr)
+mapping_process_key (Encoder *encoder, PyObject *key, PyObject **key_ptr)
 {
 	(*key_ptr) = NULL;
 	
@@ -1899,7 +1914,7 @@ mapping_process_key (JSONEncoder *encoder, PyObject *key, PyObject **key_ptr)
 }
 
 static int
-mapping_get_key_and_value_from_item (JSONEncoder *encoder, PyObject *item,
+mapping_get_key_and_value_from_item (Encoder *encoder, PyObject *item,
                                      PyObject **key_ptr, PyObject **value_ptr)
 {
 	PyObject *key = NULL, *value = NULL;
@@ -1927,7 +1942,7 @@ mapping_get_key_and_value_from_item (JSONEncoder *encoder, PyObject *item,
 
 /* Special case for dictionaries */
 static int
-write_dict (JSONEncoder *encoder, PyObject *dict, PyObject *start,
+write_dict (Encoder *encoder, PyObject *dict, PyObject *start,
             PyObject *end, PyObject *pre_value, PyObject *post_value,
             int indent_level)
 {
@@ -1980,7 +1995,7 @@ write_dict (JSONEncoder *encoder, PyObject *dict, PyObject *start,
 }
 
 static int
-write_mapping_impl (JSONEncoder *encoder, PyObject *items,
+write_mapping_impl (Encoder *encoder, PyObject *items,
                     PyObject *start, PyObject *end, PyObject *pre_value,
                     PyObject *post_value, int indent_level)
 {
@@ -2045,14 +2060,14 @@ write_mapping_impl (JSONEncoder *encoder, PyObject *items,
 }
 
 static int
-write_mapping (JSONEncoder *encoder, PyObject *mapping, int indent_level)
+write_mapping (Encoder *encoder, PyObject *mapping, int indent_level)
 {
 	int has_parents, succeeded;
 	PyObject *items;
 	PyObject *start, *end, *pre_value, *post_value;
 	
 	if (PyMapping_Size (mapping) == 0)
-		return encoder_append_const (encoder, "{}");
+		return encoder->append_ascii (encoder, "{}", 2);
 	
 	has_parents = Py_ReprEnter (mapping);
 	if (has_parents != 0)
@@ -2103,7 +2118,7 @@ write_mapping (JSONEncoder *encoder, PyObject *mapping, int indent_level)
 }
 
 static int
-check_valid_number (JSONEncoder *encoder, PyObject *serialized)
+check_valid_number (Encoder *encoder, PyObject *serialized)
 {
 	int invalid;
 	
@@ -2120,7 +2135,33 @@ check_valid_number (JSONEncoder *encoder, PyObject *serialized)
 }
 
 static PyObject *
-write_basic (JSONEncoder *encoder, PyObject *value)
+write_float (Encoder *encoder, PyObject *value)
+{
+	double val = PyFloat_AS_DOUBLE (value);
+	if (Py_IS_NAN (val))
+	{
+		PyErr_SetString (WriteError,
+				 "Cannot serialize NaN.");
+		return NULL;
+	}
+	
+	if (Py_IS_INFINITY (val))
+	{
+		const char *msg;
+		if (val > 0)
+			msg = "Cannot serialize Infinity.";
+		else
+			msg = "Cannot serialize -Infinity.";
+		
+		PyErr_SetString (WriteError, msg);
+		return NULL;
+	}
+	
+	return PyObject_Repr (value);
+}
+
+static PyObject *
+write_basic (Encoder *encoder, PyObject *value)
 {
 	if (value == Py_True)
 	{
@@ -2138,12 +2179,26 @@ write_basic (JSONEncoder *encoder, PyObject *value)
 		return encoder->null_str;
 	}
 	
+	/* Fast, exact type checks */
+	if (PyString_CheckExact (value))
+		return write_string (encoder, value);
+	if (PyUnicode_CheckExact (value))
+		return write_unicode (encoder, value);
+	if (PyInt_CheckExact (value) || PyLong_CheckExact (value))
+		return PyObject_Str (value);
+	if (PyFloat_CheckExact (value))
+		return write_float (encoder, value);
+	
+	/* Slow, full type checks */
 	if (PyString_Check (value))
 		return write_string (encoder, value);
 	if (PyUnicode_Check (value))
 		return write_unicode (encoder, value);
 	if (PyInt_Check (value) || PyLong_Check (value))
 		return PyObject_Str (value);
+	if (PyFloat_Check (value))
+		return write_float (encoder, value);
+	
 	if (PyComplex_Check (value))
 	{
 		Py_complex complex = PyComplex_AsCComplex (value);
@@ -2162,31 +2217,6 @@ write_basic (JSONEncoder *encoder, PyObject *value)
 		return NULL;
 	}
 	
-	if (PyFloat_Check (value))
-	{
-		double val = PyFloat_AS_DOUBLE (value);
-		if (Py_IS_NAN (val))
-		{
-			PyErr_SetString (WriteError,
-			                 "Cannot serialize NaN.");
-			return NULL;
-		}
-		
-		if (Py_IS_INFINITY (val))
-		{
-			const char *msg;
-			if (val > 0)
-				msg = "Cannot serialize Infinity.";
-			else
-				msg = "Cannot serialize -Infinity.";
-			
-			PyErr_SetString (WriteError, msg);
-			return NULL;
-		}
-		
-		return PyObject_Repr (value);
-	}
-	
 	if (PyObject_IsInstance (value, encoder->Decimal))
 	{
 		PyObject *serialized;
@@ -2229,7 +2259,7 @@ write_basic (JSONEncoder *encoder, PyObject *value)
 }
 
 static int
-write_object (JSONEncoder *encoder, PyObject *object,
+write_object (Encoder *encoder, PyObject *object,
               int indent_level, int in_unknown_hook)
 {
 	PyObject *pieces, *iter, *on_unknown_args;
@@ -2316,7 +2346,7 @@ valid_json_whitespace (PyObject *string)
 	
 	if (string == Py_None) return TRUE;
 	if (PyString_AsStringAndSize (string, &c_str, &c_str_len) == -1)
-		return -1;
+		return FALSE;
 	for (ii = 0; ii < (size_t)c_str_len; ii++)
 	{
 		char c = c_str[ii];
@@ -2330,7 +2360,7 @@ valid_json_whitespace (PyObject *string)
 }
 
 static int
-serializer_init_and_run_common (JSONEncoder *encoder, PyObject *value)
+serializer_init_and_run_common (Encoder *encoder, PyObject *value)
 {
 	int indent_is_valid, succeeded = FALSE;
 	
@@ -2352,20 +2382,20 @@ serializer_init_and_run_common (JSONEncoder *encoder, PyObject *value)
 	}
 	
 	if (encoder->indent_string == Py_None)
-		encoder->colon = PyString_FromString (":");
+		encoder->colon = ascii_constant (":", -1);
 	else
-		encoder->colon = PyString_FromString (": ");
+		encoder->colon = ascii_constant (": ", -1);
 	if (!encoder->colon) return FALSE;
 	
 	if ((encoder->Decimal = jsonlib_import ("decimal", "Decimal")) &&
 	    (encoder->UserString = jsonlib_import ("UserString", "UserString")) &&
-	    (encoder->true_str = unicode_from_ascii ("true")) &&
-	    (encoder->false_str = unicode_from_ascii ("false")) &&
-	    (encoder->null_str = unicode_from_ascii ("null")) &&
-	    (encoder->inf_str = unicode_from_ascii ("Infinity")) &&
-	    (encoder->neg_inf_str = unicode_from_ascii ("-Infinity")) &&
-	    (encoder->nan_str = unicode_from_ascii ("NaN")) &&
-	    (encoder->quote = unicode_from_ascii ("\"")))
+	    (encoder->true_str = ascii_constant ("true", -1)) &&
+	    (encoder->false_str = ascii_constant ("false", -1)) &&
+	    (encoder->null_str = ascii_constant ("null", -1)) &&
+	    (encoder->inf_str = ascii_constant ("Infinity", -1)) &&
+	    (encoder->neg_inf_str = ascii_constant ("-Infinity", -1)) &&
+	    (encoder->nan_str = ascii_constant ("NaN", -1)) &&
+	    (encoder->quote = ascii_constant ("\"", -1)))
 	{
 		succeeded = write_object (encoder, value, 0, FALSE);
 	}
@@ -2388,8 +2418,8 @@ static PyObject*
 _write_entry (PyObject *self, PyObject *args, PyObject *kwargs)
 {
 	PyObject *value;
-	JSONBufferEncoder encoder = {{NULL}, NULL};
-	JSONEncoder *encoder_base = (JSONEncoder*) &encoder;
+	BufferEncoder encoder = {{NULL}, NULL};
+	Encoder *encoder_base = (Encoder*) &encoder;
 	char *encoding;
 	
 	static char *kwlist[] = {"value", "sort_keys", "indent",
@@ -2415,8 +2445,8 @@ _write_entry (PyObject *self, PyObject *args, PyObject *kwargs)
 	                                  &encoder_base->on_unknown))
 		return NULL;
 	
-	encoder_base->append_ascii = encoder_buffer_append_ascii;
-	encoder_base->append_unicode = encoder_buffer_append_unicode;
+	encoder_base->append_ascii = buffer_encoder_append_ascii;
+	encoder_base->append_unicode = buffer_encoder_append_unicode;
 	if (serializer_init_and_run_common (encoder_base, value))
 	{
 		PyObject *retval;
@@ -2451,8 +2481,8 @@ static PyObject *
 _dump_entry (PyObject *self, PyObject *args, PyObject *kwargs)
 {
 	PyObject *value;
-	JSONStreamEncoder encoder = {{NULL}, NULL};
-	JSONEncoder *encoder_base = (JSONEncoder *) &encoder;
+	StreamEncoder encoder = {{NULL}, NULL};
+	Encoder *encoder_base = (Encoder *) &encoder;
 	
 	static char *kwlist[] = {"value", "fp", "sort_keys", "indent",
 	                         "ascii_only", "coerce_keys", "encoding",
@@ -2478,8 +2508,8 @@ _dump_entry (PyObject *self, PyObject *args, PyObject *kwargs)
 	                                  &encoder_base->on_unknown))
 		return NULL;
 	
-	encoder_base->append_ascii = encoder_stream_append_ascii;
-	encoder_base->append_unicode = encoder_stream_append_unicode;
+	encoder_base->append_ascii = stream_encoder_append_ascii;
+	encoder_base->append_unicode = stream_encoder_append_unicode;
 	if (serializer_init_and_run_common (encoder_base, value))
 	{
 		Py_INCREF (Py_None);
@@ -2612,7 +2642,7 @@ initjsonlib (void)
 	/* If you change the version here, also change it in setup.py and
 	 * jsonlib.py.
 	**/
-	version = Py_BuildValue ("(iii)", 1, 3, 8);
+	version = Py_BuildValue ("(iii)", 1, 3, 9);
 	PyModule_AddObject (module, "__version__", version);
 }
 /* }}} */
